using System;
using System.Windows.Forms;

using System.Collections;
using System.Threading;
using System.Drawing;
using Flock;
using Vector;

using Steema.TeeChart;
using Steema.TeeChart.Styles;

// === File Prolog ============================================================
// This code was developed by RiverSoftAVG.
// --- Notes ------------------------------------------------------------------
// This flocking demo borrows concepts and algorithms from Steven Woodcock's
// article, "Flocking: A Simple Technique for Simulating Group Behavior,"
// which is found in the book, Game Programming Gems

// === End File Prolog ========================================================
namespace FlockMain
{
    public partial class TfrmMain : Form
    {
        private double MaxTimeOfStep
        {
            get
            {
                return FMaxTimeOfStep;
            }
        }
        // Public declarations
        public bool Simulate
        {
            get
            {
                return FSimulate;
            }
        }
        public int NumShapes
        {
            get
            {
                return FNumShapes;
            }
        }
        // Private declarations
        private double FMaxTimeOfStep = 0;
        private bool FSimulate = false;
        private int FNumShapes = 0;
        public TfrmMain()
        {
            if (ProtoFlock == null)
                this.ProtoFlock = new Flock.TFlock(null);
            /***************/
            InitializeComponent();
            /***************/
            this.ProtoFlock.OnChange = new TNotifyEvent(ProtoFlockChange);

            this.ProtoFlock.OnBoidChange = new TBoidEvent(ProtoFlockApplyBoidChange);
            this.ProtoFlock.OnBoidAdded = new TBoidEvent(ProtoFlockApplyBoidAdded);
            this.ProtoFlock.OnBoidDelete = new TBoidEvent(ProtoFlockApplyBoidDelete);
            this.ProtoFlock.OnBoidExecuted = new TBoidEvent(ProtoFlockApplyBoidExecuted);
            this.ProtoFlock.OnBoidExecuting = new TBoidEvent(ProtoFlockApplyBoidExecuting);

            this.ProtoFlock.OnApplyAlignment = new TBoidApplyChangeEvent(ProtoFlockApplyAlignment);
            this.ProtoFlock.OnApplyAvoidance = new TBoidApplyChangeEvent(ProtoFlockApplyAvoidance);
            this.ProtoFlock.OnApplyCohesion = new TBoidApplyChangeEvent(ProtoFlockApplyCohesion);
            this.ProtoFlock.OnApplySeparation = new TBoidApplyChangeEvent(ProtoFlockApplySeparation);


        }

        public void FormCreate(System.Object Sender, System.EventArgs _e1)
        {
            int i;
            // Figure out the number of shape series
            FNumShapes = 0;

            for (i = 0; i < Chart1.Series.Count; i++)
            {

                if (Chart1.Series[i] is Steema.TeeChart.Styles.Shape)
                {
                    if (i > NumShapes)
                    {
                        // Non-shape in-between, move shape to front

                        Chart1.Series.Exchange(NumShapes, i);
                    }
                    FNumShapes++;
                }
            }
            // Set Chart based on world

            Chart1.Axes.Left.Minimum = World.Origin.Z;

            Chart1.Axes.Left.Maximum = World.Origin.Z + World.Dimension.Z;

            Chart1.Axes.Bottom.Minimum = World.Origin.X;

            Chart1.Axes.Bottom.Maximum = World.Origin.X + World.Dimension.X;
            // set the maximum time a step may take
            FMaxTimeOfStep = 1000 / tbMaxStepsPerSec.Value;
            // Use the ProtoFlock to set controls
            for (i = 0; i < clbBehaviors.Items.Count; i++)
            {

                this.clbBehaviors.SetItemChecked(i, true);
                ProtoFlock.Behaviors[i] = (TFlockingBehavior)(i + 1);

            }
            tbAlignmentStrength.Value = (int)Math.Round(ProtoFlock.StrengthAlignment * 100f);
            tbAvoidanceStrength.Value = (int)Math.Round(ProtoFlock.StrengthAvoidance * 100);
            tbSeparationStrength.Value = (int)Math.Round(ProtoFlock.StrengthSeparation * 100);
            tbCohesionStrength.Value = (int)Math.Round(ProtoFlock.StrengthCohesion * 100);
            tbNumFlocksChange(tbNumFlocks, _e1);
        }

        public void btnOneStepClick(System.Object Sender, System.EventArgs _e1)
        {
            // Update the flock one step
            UpdateFlock();
        }

        public void btnStartClick(System.Object Sender, System.EventArgs _e1)
        {
            double Time;
            double WaitTime;
            // Set flag that controls whether the simulation terminates
            FSimulate = true;
            // enable the stop button to allow user to stop simulation
            btnStart.Enabled = false;
            btnStop.Enabled = true;
            // Every time step, update the flock and allow the UI to process
            // messages so that user can stop the simulation
            while (Simulate)
            {

                Time = System.DateTime.Now.Ticks/*TimeGetTime*/;
                UpdateFlock();

                Application.DoEvents()/*ProcessMessages*/;

                WaitTime = System.DateTime.Now.Ticks - Time;
                WaitTime = MaxTimeOfStep - WaitTime;
                if (WaitTime > 0)
                {
                    Thread.Sleep((int)WaitTime);
                }

                Application.DoEvents()/*ProcessMessages*/;
            }
        }

        public void btnStopClick(System.Object Sender, System.EventArgs _e1)
        {
            // Stop simulation and enable Start button again
            FSimulate = false;
            btnStart.Enabled = true;
            btnStop.Enabled = false;
        }

        // Note: the original parameters are Object Sender, ref bool CanClose
        public void FormCloseQuery(System.Object Sender, System.ComponentModel.CancelEventArgs _e1)
        {
            // Stop simulating if we need to
            FSimulate = false;
            _e1.Cancel/*CanClose*/ = true;

        }

        private void UpdateFlock()
        {
            int i;
            int j;
            TCustomFlock Flock;
            for (i = 0; i < World.FlockCount; i++)
            {

                Flock = World.Flock(i);
                // Update the flock... apply the flocking rules
                Flock.Execute();
                // Zero out any Y components that may have been introduced since
                // this is a 2D simulation
                for (j = 0; j < Flock.Boids.Count; j++)
                {

                    Flock.Boids[j].Position.Y = 0f;

                    Flock.Boids[j].Orientation.Y = 0f;
                }
            }
        }

        public void clbBehaviorsClickCheck(Object Sender, System.EventArgs _e1)
        {
            TFlockingBehavior[] Behaviors;
            int i;
            // Change the behaviors of the flocks
            Behaviors = new TFlockingBehavior[4] { TFlockingBehavior.none, TFlockingBehavior.none, TFlockingBehavior.none, TFlockingBehavior.none };

            if (clbBehaviors.GetItemChecked(0))
            {

                Behaviors[0] = TFlockingBehavior.fbSeparation;

            }

            if (clbBehaviors.GetItemChecked(1))
            {

                Behaviors[1] = TFlockingBehavior.fbAlignment;
            }

            if (clbBehaviors.GetItemChecked(2))
            {

                Behaviors[2] = TFlockingBehavior.fbCohesion;
            }

            if (clbBehaviors.GetItemChecked(3))
            {

                Behaviors[3] = TFlockingBehavior.fbAvoidance;
            }
            ProtoFlock.Behaviors = Behaviors;
            for (i = 0; i < World.FlockCount; i++)
            {

                World.Flock(i).Behaviors = Behaviors;
            }
        }

        public void tbMaxStepsPerSecChange(System.Object Sender, System.EventArgs _e1)
        {
            // set the maximum time a step can take based on the number of
            // steps the user wants every second
            FMaxTimeOfStep = (double)1000 / ((TrackBar)Sender).Value;
        }

        private void CreateFlock()
        {
            // Create the flock and populate it
            TFlock _wvar1 = new TFlock(World);
            _wvar1.Assign(ProtoFlock);
            //_wvar1 = ProtoFlock;
            _wvar1.World = this.World;
            // Hook up the flock's OnChange event so that the flock automatically
            // informs the TeeChart of any changes
            _wvar1.OnChange = ProtoFlock.OnChange;
            _wvar1.OnApplyAvoidance = ProtoFlock.OnApplyAvoidance;//TBoidApplyChangeEvent

            // Store in the flock's tag a pointer to its Arrow series

            Arrow arrow = new Arrow(Chart1.Chart);
            _wvar1.Tag = arrow;


            //Tag.Chart = (Chart)Chart1;//sr


            // create the boids
            tbBoidsPerFlockChange(tbBoidsPerFlock, null);//sr
        }

        public void tbBoidsPerFlockChange(System.Object Sender, System.EventArgs _e1)
        {
            int i;
            int j;
            TCustomFlock Flock;
            System.Random tmp1 = new System.Random(100);
            System.Random tmp2 = new System.Random(2);


            for (i = 0; i < World.FlockCount; i++)
            {

                Flock = World.Flock(i);
                // This procedure changes the number of boids in each flock

                Flock.BeginUpdate();
                try
                {
                    TrackBar _wvar1 = (TrackBar)Sender;
                    if (_wvar1.Value < Flock.Boids.Count)
                    {
                        int delta = Flock.Boids.Count - _wvar1.Value;

                        for (j = 0; j < delta; j++)
                        {
                            Flock.Boids.Delete(0);
                        }
                    }
                    else if (_wvar1.Value > Flock.Boids.Count)
                    {
                        int delta = _wvar1.Value - Flock.Boids.Count;

                        for (j = 0; j < delta; j++)
                        {
                            TBoid _wvar2 = Flock.Boids.Add();

                            float t1 = (float)tmp1.NextDouble() * 100;
                            float t2 = (float)tmp1.NextDouble() * 100;
                            _wvar2.Position.SetXYZ(t1, 0f, t2);
                            t1 = (float)tmp2.NextDouble() * 2;
                            t2 = (float)tmp2.NextDouble() * 2;
                            _wvar2.Orientation.SetXYZ(t1 - 1, 0, t2 - 1);
                            _wvar2.Orientation.Normalize();
                        }
                    }
                }
                finally
                {

                    Flock.EndUpdate();// Boids.EndUpdate;
                }
            }
        }

        public void tbNumFlocksChange(System.Object Sender, System.EventArgs _e1)
        {
            int j;
            TrackBar _wvar1 = (TrackBar)Sender;
            if (_wvar1.Value < World.FlockCount)
            {
                for (j = 0; j <= (World.FlockCount - 1 - _wvar1.Value); j++)
                {
                    // Clear the series and then free the flock

                    ((Arrow)(World.Flock(World.FlockCount - 1).Tag)).Clear();

                    World.Flocks.RemoveAt(World.FlockCount - 1);
                }
            }
            else if (_wvar1.Value > World.FlockCount)
            {
                for (j = 0; j <= (_wvar1.Value - (World.FlockCount + 1)); j++)
                {
                    CreateFlock();
                }
            }
        }

        public void tbSeparationStrengthChange(System.Object Sender, System.EventArgs _e1)
        {
            int i;
            TrackBar _wvar1 = (TrackBar)Sender;
            if (Sender == tbSeparationStrength)
            {
                ProtoFlock.StrengthSeparation = (float)_wvar1.Value / 100;
                for (i = 0; i < World.FlockCount; i++)
                {

                    World.Flock(i).StrengthSeparation = (float)_wvar1.Value / 100;
                }
            }
            else if (Sender == tbAlignmentStrength)
            {
                ProtoFlock.StrengthAlignment = (float)_wvar1.Value / 100;
                for (i = 0; i < World.FlockCount; i++)
                {

                    World.Flock(i).StrengthAlignment = (float)_wvar1.Value / 100;
                }
            }
            else if (Sender == tbCohesionStrength)
            {
                ProtoFlock.StrengthCohesion = (float)_wvar1.Value / 100;
                for (i = 0; i < World.FlockCount; i++)
                {

                    World.Flock(i).StrengthCohesion = (float)_wvar1.Value / 100;
                }
            }
            else if (Sender == tbAvoidanceStrength)
            {
                ProtoFlock.StrengthAvoidance = (float)_wvar1.Value / 100;
                for (i = 0; i < World.FlockCount; i++)
                {

                    World.Flock(i).StrengthAvoidance = (float)_wvar1.Value / 100;
                }
            }


            else if (Sender == tbNeigboursRank)
            {
                ProtoFlock.NeigboursRankSeparation = (float)_wvar1.Value;
                for (i = 0; i < World.FlockCount; i++)
                {

                    World.Flock(i).NeigboursRankSeparation = (float)_wvar1.Value;
                }
            }

            else if (Sender == tbResponseDelay)
            {
                ProtoFlock.responseDelay = _wvar1.Value;
                for (i = 0; i < World.FlockCount; i++)
                {

                    World.Flock(i).responseDelay = _wvar1.Value;
                }
            }

            else if (Sender == tbMaxSpeed)
            {
                ProtoFlock.MaxSpeed = (float)_wvar1.Value / 100;
                for (i = 0; i < World.FlockCount; i++)
                {

                    World.Flock(i).MaxSpeed = (float)_wvar1.Value / 100;
                }
            }

        }

        public void Chart1MouseMove(Object Sender, Keys Shift, int X, int Y)
        {
            double NewX;
            double NewY;

            if (!(new ArrayList(World.Flock(0).AutoCalcProperties).Contains(Flock.TAutoCalcProperty.acpPosition)))
            {


                NewX = X;//((TChart)Sender).Series[NumShapes].;//GetCursorValues(NewX, NewY);
                NewY = Y;


                World.Flock(0)/*[0]*/.Position.SetXYZ((float)NewX, 0f, (float)NewY);
            }
        }

        public void cbFollowMouseClick(System.Object Sender, System.EventArgs _e1)
        {
            if (((CheckBox)Sender).Checked)
            {

                World.Flock(0).AutoCalcProperties = new TAutoCalcProperty[2] { Flock.TAutoCalcProperty.acpVelocity, Flock.TAutoCalcProperty.acpVelocity };
            }
            else
            {

                World.Flock(0).AutoCalcProperties = new TAutoCalcProperty[] { Flock.TAutoCalcProperty.acpPosition, Flock.TAutoCalcProperty.acpVelocity };
            }
        }

        public void ProtoFlockChange(Object Sender)
        {
            int i;
            Arrow/*TArrowSeries*/ AFlockSeries;
            // Every time the flock changes, redraw everything
            TFlock _wvar1 = (TFlock)Sender;
            if (Sender == ProtoFlock)
            {
                return;
            }

            AFlockSeries = /*_wvar1.TArrowSeries*/(Arrow)(_wvar1.Tag);

            AFlockSeries.Clear();

            //System.Drawing.Color cl= new System.Drawing.Color();
            //cl = System.Drawing.Color.Black;

            for (i = 0; i < _wvar1.Boids.Count; i++)
            {

                AFlockSeries.Add/*AddArrow*/(_wvar1.Boids[i].Position.X, _wvar1.Boids[i].Position.Z, _wvar1.Boids[i].Position.X + _wvar1.Boids[i].Velocity.X * 2 + 0.1, _wvar1.Boids[i].Position.Z + _wvar1.Boids[i].Velocity.Z * 2 + 0.1, ""/*, cl_wvar1.clTeeColor*/);
            }
        }

        public void ProtoFlockApplyBoidAdded(Object Sender, TBoid Boid)
        {
            ProtoFlockChange(Boid.Flock);

        }
        public void ProtoFlockApplyBoidChange(Object Sender, TBoid Boid)
        {
            ProtoFlockChange(Boid.Flock);

        }
        public void ProtoFlockApplyBoidDelete(Object Sender, TBoid Boid)
        {
            ProtoFlockChange(Boid.Flock);

        }
        public void ProtoFlockApplyBoidExecuted(Object Sender, TBoid Boid)
        {
            ProtoFlockChange(Boid.Flock);

        }

        public void ProtoFlockApplyBoidExecuting(Object Sender, TBoid Boid)
        {
            ProtoFlockChange(Boid.Flock);

        }




        public void ProtoFlockApplyCohesion(Object Sender, TBoid Boid, TVector Accumulator, TVector Change)
        {


        }

        public void ProtoFlockApplyAlignment(Object Sender, TBoid Boid, TVector Accumulator, TVector Change)
        {

        }

        public void ProtoFlockApplySeparation(Object Sender, TBoid Boid, TVector Accumulator, TVector Change)
        {


        }


        public void ProtoFlockApplyAvoidance(Object Sender, TBoid Boid, TVector Accumulator, TVector Change)
        {
            int i;
            TVector v;
            double Radius;
            // make sure we avoid the shapes too!
            // In a real application, you would probably move the avoidance into
            // a TFlockWorld or TBoid descendant
            // Note: this routine uses a collision sphere to detect collisions
            // we also make an assumption that two obstacles are not close enough to
            // set off collision avoidance at the same time so we can quit when we
            // detect the first imminent collision
            v = new TVector();
            try
            {
                // we signal a collision by setting the tag of the boid to the shape index
                // we collided with
                //Boid.Tag =  -1;
                // find nearest shape
                for (i = 0; i < NumShapes; i++)
                {
                    // calculate center point

                    v.X = (float)((((Shape)Chart1.Series[i]).X1 - ((Shape)Chart1.Series[i]).X0) / 2 + ((Shape)Chart1.Series[i]).X0);

                    v.Z = (float)((((Shape)Chart1.Series[i]).Y1 - ((Shape)Chart1.Series[i]).Y0) / 2 + ((Shape)Chart1.Series[i]).Y0);




                    Radius = Math.Max(Math.Sqrt(Math.Pow(((Shape)Chart1.Series[i]).X0 - v.X, 2) + Math.Pow(((Shape)Chart1.Series[i]).Y0 - v.Z, 2)),
                                      Math.Sqrt(Math.Pow(((Shape)Chart1.Series[i]).X1 - v.X, 2) + Math.Pow(((Shape)Chart1.Series[i]).Y0 - v.Z, 2)));

                    double tmp = Math.Max(Math.Sqrt(Math.Pow(((Shape)Chart1.Series[i]).X0 - v.X, 2) + Math.Pow(((Shape)Chart1.Series[i]).Y1 - v.Z, 2)),
                                          Math.Sqrt(Math.Pow(((Shape)Chart1.Series[i]).X1 - v.X, 2) + Math.Pow(((Shape)Chart1.Series[i]).Y1 - v.Z, 2)));

                    Radius = Math.Max(Radius, tmp);


                    if (v.DistanceTo(Boid.Position) < (Radius + 6 * Boid.Flock.StrengthAvoidance)) // I change from 2 * SeparationDistance
                    {
                        // Move AWAY NOW!
                        // we are going to override all accumulations
                        // we reset the change to move directly away... to be more
                        // realistic, they should probably move along the wall of the
                        // shape
                        Change.Assign(Boid.Position);
                        Change.Subtract(v);

                        Change.SetMagnitudeOfVector(Math.Max(Boid.Flock.MaxUrgency, 1));
                        //Boid.Tag = 1;
                        return;
                    }
                }
            }
            finally
            {

                // v.Free;
            }
        }

        private void Chart1MouseMove(object sender, MouseEventArgs e)
        {
            double NewX;
            double NewY;

            if (World.Flock(0) != null)
            {
                Point p = new Point(e.X, e.Y);

                //Point p = ((TChart)sender).Cursor.HotSpot.X;// .Series[NumShapes].;//GetCursorValues(NewX, NewY);

                NewX = Math.Min(100, ((double)p.X / 600.0f) * 100.0f - 8.8);

                NewY = Math.Min(100, (1 - (double)(p.Y / 600.0f)) * 100.0f + 5);
                //NewY = ((TChart)sender).Cursor.HotSpot.Y;

                if ((World.Flock(0).AutoCalcProperties[0] != (TAutoCalcProperty.acpPosition)))
                    World.Flock(0).Position.SetXYZ((float)NewX, 0f, (float)NewY);


            }

        }

        private void checkBox1_CheckedChanged(object sender, EventArgs e)
        {

            if (((CheckBox)sender).Checked)
            {

                World.Flock(0).localInformation = true;
            }
            else
                World.Flock(0).localInformation = false;

        }


    } // end TfrmMain

}

namespace FlockMain.Units
{
    public class FlockMain
    {
        public static TfrmMain frmMain = null;
    } // end FlockMain

}
