using System;

// === File Prolog ============================================================

// === End File Prolog ========================================================
namespace Vector
{
    public class TVector : Object
    {
        // Published declarations
        public float X
        {
            get
            {
                return FX;
            }
            set
            {
                SetX(value);
            }
        }
        public float Y
        {
            get
            {
                return FY;
            }
            set
            {
                SetY(value);
            }
        }
        public float Z
        {
            get
            {
                return FZ;
            }
            set
            {
                SetZ(value);
            }
        }
        // Private declarations
        private float FX = 0;
        private float FY = 0;
        private float FZ = 0;
        // TVector
        // TVector
        public void Add(TVector Vector)
        {
            FX = FX + Vector.X;
            FY = FY + Vector.Y;
            FZ = FZ + Vector.Z;
            Change();
        }

        public void Add(float Value)
        {
            FX = FX + Value;
            FY = FY + Value;
            FZ = FZ + Value;
            Change();
        }

        public /*override*/ void Assign(Object Source)
        {
            if (Source is TVector)
            {
                FX = ((Source) as TVector).X;
                FY = ((Source) as TVector).Y;
                FZ = ((Source) as TVector).Z;
                Change();
                return;
            }
            //base.Assign(Source);
        }

        public virtual void Clear()
        {
            FX = 0;
            FY = 0;
            FZ = 0;
            Change();
        }

        //Constructor  Create( Vector)
        public TVector(TVector Vector)
            : base()
        {
            FX = Vector.X;
            FY = Vector.Y;
            FZ = Vector.Z;
        }
        //Constructor  Create( X,  Y,  Z)
        public TVector(float X, float Y, float Z)
            : base()
        {
            FX = X;
            FY = Y;
            FZ = Z;
        }
        //Constructor  Create()
        public TVector()
            : base()
        {
        }
        public void CrossProduct(TVector Vector)
        {
            float tempX;
            float tempY;
            float tempZ;
            tempX = X * Vector.Z - Z * Vector.Y;
            tempY = Z * Vector.X - X * Vector.Z;
            tempZ = X * Vector.Y - Y * Vector.X;
            SetXYZ(tempX, tempY, tempZ);
        }

        public void CrossProduct(float X, float Y, float Z)
        {
            float tempX;
            float tempY;
            float tempZ;
            tempX = FX * Z - FZ * Y;
            tempY = FZ * X - FX * Z;
            tempZ = FX * Y - FY * X;
            SetXYZ(tempX, tempY, tempZ);
        }

        public float DistanceTo(TVector Vector)
        {
            float result;
            result = (float)Math.Sqrt(Math.Pow((X - Vector.X), 2) + Math.Pow((Y - Vector.Y), 2) + Math.Pow((Z - Vector.Z), 2));
            return result;
        }

        public double DistanceSqrTo(TVector Vector)
        {
            double result;
            result = Math.Pow(X - Vector.X, 2) + Math.Pow(Y - Vector.Y, 2) + Math.Pow(Z - Vector.Z, 2);
            return result;
        }

        public double DistanceSqrTo(float X, float Y, float Z)
        {
            double result;
            result = Math.Pow(FX - X, 2) + Math.Pow(FY - Y, 2) + Math.Pow(FZ - Z, 2);
            return result;
        }

        public float DistanceTo(float X, float Y, float Z)
        {
            float result;
            result = (float)Math.Sqrt(Math.Pow(FX - X, 2) + Math.Pow(FY - Y, 2) + Math.Pow(FZ - Z, 2));
            return result;
        }

        public void Divide(TVector Vector)
        {
            FX = X / Vector.X;
            FY = Y / Vector.Y;
            FZ = Z / Vector.Z;
            Change();
        }

        public void Divide(float Value)
        {
            FX = X / Value;
            FY = Y / Value;
            FZ = Z / Value;
            Change();
        }

        public float DotProduct(TVector Vector)
        {
            float result;
            result = (X * Vector.X) + (Y * Vector.Y) + (Z * Vector.Z);
            return result;
        }

        public float DotProduct(float X, float Y, float Z)
        {
            float result;
            result = (FX * X) + (FY * Y) + (FZ * Z);
            return result;
        }

        public bool Equals(float X, float Y, float Z)
        {
            bool result;
            result = (FX >= (X - Units.Vector.FloatError)) && (FX <= (X + Units.Vector.FloatError)) && (FY >= (Y - Units.Vector.FloatError)) && (FY <= (Y + Units.Vector.FloatError)) && (FZ >= (Z - Units.Vector.FloatError)) && (FZ <= (Z + Units.Vector.FloatError));
            return result;
        }

        public bool Equals(TVector Vector)
        {
            bool result;
            result = (FX >= (Vector.X - Units.Vector.FloatError)) && (FX <= (Vector.X + Units.Vector.FloatError)) && (FY >= (Vector.Y - Units.Vector.FloatError)) && (FY <= (Vector.Y + Units.Vector.FloatError)) && (FZ >= (Vector.Z - Units.Vector.FloatError)) && (FZ <= (Vector.Z + Units.Vector.FloatError));
            return result;
        }

        public bool EqualsStar(float X, float Y, float Z)
        {
            bool result;
            TVector v1;
            TVector v2;
            // see if vectors point in same direction
            v1 = new TVector(X, Y, Z);
            v2 = new TVector(this);
            try
            {
                v1.Normalize();
                v2.Normalize();
                result = v1.Equals(v2);
            }
            finally
            {

                //v1.Free;

                //v2.Free;
            }
            return result;
        }

        public bool EqualsStar(TVector Vector)
        {
            bool result;
            TVector v1;
            TVector v2;
            // see if vectors point in same direction
            v1 = new TVector(Vector);
            v2 = new TVector(this);
            try
            {
                v1.Normalize();
                v2.Normalize();
                result = v1.Equals(v2);
            }
            finally
            {

                //v1.Free;

                //v2.Free;
            }
            return result;
        }

        public float Length()
        {
            float result;
            result = (float)Math.Sqrt(X * X + Y * Y + Z * Z);
            return result;
        }

        public void Multiply(TVector Vector)
        {
            FX = X * Vector.X;
            FY = Y * Vector.Y;
            FZ = Z * Vector.Z;
            Change();
        }

        public void Multiply(float Value)
        {
            FX = X * Value;
            FY = Y * Value;
            FZ = Z * Value;
            Change();
        }

        public void Normalize()
        {
            double m;
            m = Length();
            if (m > 0)
            {
                m = 1 / m;
            }
            else
            {
                m = 0;
            }
            FX = (float)(X * m);
            FY = (float)(Y * m);
            FZ = (float)(Z * m);
            Change();
        }

        public void SetMagnitudeOfVector(float Velocity)
        {
            Normalize();
            FX = X * Velocity;
            FY = Y * Velocity;
            FZ = Z * Velocity;
            Change();
        }

        public void SetXYZ(float X, float Y, float Z)
        {
            FX = X;
            FY = Y;
            FZ = Z;
            Change();
        }

        public void Subtract(TVector Vector)
        {
            FX = FX - Vector.X;
            FY = FY - Vector.Y;
            FZ = FZ - Vector.Z;
            Change();
        }

        public void Subtract(float Value)
        {
            FX = FX - Value;
            FY = FY - Value;
            FZ = FZ - Value;
            Change();
        }

        // Protected declarations
        public virtual void SetX(float Value)
        {
            FX = Value;
            Change();
        }

        public virtual void SetY(float Value)
        {
            FY = Value;
            Change();
        }

        public virtual void SetZ(float Value)
        {
            FZ = Value;
            Change();
        }

        public virtual void Change()
        {
        }

    } // end TVector

}

namespace Vector.Units
{
    public class Vector
    {
        public const double FloatError = 0.001;
    } // end Vector

}




